# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: example2.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='example2.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0e\x65xample2.proto\"5\n\x15\x45xampleReferencedType\x12\r\n\x05key_1\x18\x01 \x01(\t\x12\r\n\x05key_2\x18\x02 \x01(\t*5\n\x15\x45xampleReferencedEnum\x12\x08\n\x04KEY0\x10\x00\x12\x08\n\x04KEY1\x10\x01\x12\x08\n\x04KEY2\x10\x02\x62\x06proto3')
)

_EXAMPLEREFERENCEDENUM = _descriptor.EnumDescriptor(
  name='ExampleReferencedEnum',
  full_name='ExampleReferencedEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='KEY0', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KEY1', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='KEY2', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=73,
  serialized_end=126,
)
_sym_db.RegisterEnumDescriptor(_EXAMPLEREFERENCEDENUM)

ExampleReferencedEnum = enum_type_wrapper.EnumTypeWrapper(_EXAMPLEREFERENCEDENUM)
KEY0 = 0
KEY1 = 1
KEY2 = 2



_EXAMPLEREFERENCEDTYPE = _descriptor.Descriptor(
  name='ExampleReferencedType',
  full_name='ExampleReferencedType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_1', full_name='ExampleReferencedType.key_1', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key_2', full_name='ExampleReferencedType.key_2', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18,
  serialized_end=71,
)

DESCRIPTOR.message_types_by_name['ExampleReferencedType'] = _EXAMPLEREFERENCEDTYPE
DESCRIPTOR.enum_types_by_name['ExampleReferencedEnum'] = _EXAMPLEREFERENCEDENUM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ExampleReferencedType = _reflection.GeneratedProtocolMessageType('ExampleReferencedType', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLEREFERENCEDTYPE,
  '__module__' : 'example2_pb2'
  # @@protoc_insertion_point(class_scope:ExampleReferencedType)
  })
_sym_db.RegisterMessage(ExampleReferencedType)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: compat/example_compat.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='compat/example_compat.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1b\x63ompat/example_compat.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"l\n\x18\x45xampleCompatNestedModel\x12\x12\n\nstring_key\x18\x01 \x01(\t\x12\x11\n\tint32_key\x18\x02 \x01(\x05\x12)\n\x08\x65num_key\x18\x03 \x01(\x0e\x32\x17.ExampleCompatEnumModel\"\xc2\x04\n\x14\x45xampleCompatDBModel\x12\x11\n\tint32_key\x18\x01 \x01(\x05\x12\x12\n\nstring_key\x18\x02 \x01(\t\x12\x10\n\x08\x62ool_key\x18\x03 \x01(\x08\x12\x11\n\tbytes_key\x18\x04 \x01(\x0c\x12\x12\n\ndouble_key\x18\x0e \x01(\x01\x12\x11\n\tint64_key\x18\x10 \x01(\x03\x12\x45\n\x11map_string_string\x18\x05 \x03(\x0b\x32*.ExampleCompatDBModel.MapStringStringEntry\x12\x43\n\x10map_string_int32\x18\x06 \x03(\x0b\x32).ExampleCompatDBModel.MapStringInt32Entry\x12\x18\n\x10string_array_key\x18\x07 \x03(\t\x12\x17\n\x0fint32_array_key\x18\x08 \x03(\x05\x12)\n\x08\x65num_key\x18\n \x01(\x0e\x32\x17.ExampleCompatEnumModel\x12\x31\n\rtimestamp_key\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12+\n\nstruct_key\x18\x0c \x01(\x0b\x32\x17.google.protobuf.Struct\x1a\x36\n\x14MapStringStringEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x35\n\x13MapStringInt32Entry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01*<\n\x16\x45xampleCompatEnumModel\x12\n\n\x06\x45NUM10\x10\x00\x12\n\n\x06\x45NUM11\x10\x01\x12\n\n\x06\x45NUM12\x10\x02\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])

_EXAMPLECOMPATENUMMODEL = _descriptor.EnumDescriptor(
  name='ExampleCompatEnumModel',
  full_name='ExampleCompatEnumModel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ENUM10', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENUM11', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENUM12', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=785,
  serialized_end=845,
)
_sym_db.RegisterEnumDescriptor(_EXAMPLECOMPATENUMMODEL)

ExampleCompatEnumModel = enum_type_wrapper.EnumTypeWrapper(_EXAMPLECOMPATENUMMODEL)
ENUM10 = 0
ENUM11 = 1
ENUM12 = 2



_EXAMPLECOMPATNESTEDMODEL = _descriptor.Descriptor(
  name='ExampleCompatNestedModel',
  full_name='ExampleCompatNestedModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='string_key', full_name='ExampleCompatNestedModel.string_key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_key', full_name='ExampleCompatNestedModel.int32_key', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='enum_key', full_name='ExampleCompatNestedModel.enum_key', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=94,
  serialized_end=202,
)


_EXAMPLECOMPATDBMODEL_MAPSTRINGSTRINGENTRY = _descriptor.Descriptor(
  name='MapStringStringEntry',
  full_name='ExampleCompatDBModel.MapStringStringEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ExampleCompatDBModel.MapStringStringEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ExampleCompatDBModel.MapStringStringEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=674,
  serialized_end=728,
)

_EXAMPLECOMPATDBMODEL_MAPSTRINGINT32ENTRY = _descriptor.Descriptor(
  name='MapStringInt32Entry',
  full_name='ExampleCompatDBModel.MapStringInt32Entry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ExampleCompatDBModel.MapStringInt32Entry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ExampleCompatDBModel.MapStringInt32Entry.value', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=730,
  serialized_end=783,
)

_EXAMPLECOMPATDBMODEL = _descriptor.Descriptor(
  name='ExampleCompatDBModel',
  full_name='ExampleCompatDBModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='int32_key', full_name='ExampleCompatDBModel.int32_key', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key', full_name='ExampleCompatDBModel.string_key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bool_key', full_name='ExampleCompatDBModel.bool_key', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bytes_key', full_name='ExampleCompatDBModel.bytes_key', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='double_key', full_name='ExampleCompatDBModel.double_key', index=4,
      number=14, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int64_key', full_name='ExampleCompatDBModel.int64_key', index=5,
      number=16, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='map_string_string', full_name='ExampleCompatDBModel.map_string_string', index=6,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='map_string_int32', full_name='ExampleCompatDBModel.map_string_int32', index=7,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_array_key', full_name='ExampleCompatDBModel.string_array_key', index=8,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_array_key', full_name='ExampleCompatDBModel.int32_array_key', index=9,
      number=8, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='enum_key', full_name='ExampleCompatDBModel.enum_key', index=10,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp_key', full_name='ExampleCompatDBModel.timestamp_key', index=11,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='struct_key', full_name='ExampleCompatDBModel.struct_key', index=12,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_EXAMPLECOMPATDBMODEL_MAPSTRINGSTRINGENTRY, _EXAMPLECOMPATDBMODEL_MAPSTRINGINT32ENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=205,
  serialized_end=783,
)

_EXAMPLECOMPATNESTEDMODEL.fields_by_name['enum_key'].enum_type = _EXAMPLECOMPATENUMMODEL
_EXAMPLECOMPATDBMODEL_MAPSTRINGSTRINGENTRY.containing_type = _EXAMPLECOMPATDBMODEL
_EXAMPLECOMPATDBMODEL_MAPSTRINGINT32ENTRY.containing_type = _EXAMPLECOMPATDBMODEL
_EXAMPLECOMPATDBMODEL.fields_by_name['map_string_string'].message_type = _EXAMPLECOMPATDBMODEL_MAPSTRINGSTRINGENTRY
_EXAMPLECOMPATDBMODEL.fields_by_name['map_string_int32'].message_type = _EXAMPLECOMPATDBMODEL_MAPSTRINGINT32ENTRY
_EXAMPLECOMPATDBMODEL.fields_by_name['enum_key'].enum_type = _EXAMPLECOMPATENUMMODEL
_EXAMPLECOMPATDBMODEL.fields_by_name['timestamp_key'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXAMPLECOMPATDBMODEL.fields_by_name['struct_key'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
DESCRIPTOR.message_types_by_name['ExampleCompatNestedModel'] = _EXAMPLECOMPATNESTEDMODEL
DESCRIPTOR.message_types_by_name['ExampleCompatDBModel'] = _EXAMPLECOMPATDBMODEL
DESCRIPTOR.enum_types_by_name['ExampleCompatEnumModel'] = _EXAMPLECOMPATENUMMODEL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ExampleCompatNestedModel = _reflection.GeneratedProtocolMessageType('ExampleCompatNestedModel', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLECOMPATNESTEDMODEL,
  '__module__' : 'compat.example_compat_pb2'
  # @@protoc_insertion_point(class_scope:ExampleCompatNestedModel)
  })
_sym_db.RegisterMessage(ExampleCompatNestedModel)

ExampleCompatDBModel = _reflection.GeneratedProtocolMessageType('ExampleCompatDBModel', (_message.Message,), {

  'MapStringStringEntry' : _reflection.GeneratedProtocolMessageType('MapStringStringEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXAMPLECOMPATDBMODEL_MAPSTRINGSTRINGENTRY,
    '__module__' : 'compat.example_compat_pb2'
    # @@protoc_insertion_point(class_scope:ExampleCompatDBModel.MapStringStringEntry)
    })
  ,

  'MapStringInt32Entry' : _reflection.GeneratedProtocolMessageType('MapStringInt32Entry', (_message.Message,), {
    'DESCRIPTOR' : _EXAMPLECOMPATDBMODEL_MAPSTRINGINT32ENTRY,
    '__module__' : 'compat.example_compat_pb2'
    # @@protoc_insertion_point(class_scope:ExampleCompatDBModel.MapStringInt32Entry)
    })
  ,
  'DESCRIPTOR' : _EXAMPLECOMPATDBMODEL,
  '__module__' : 'compat.example_compat_pb2'
  # @@protoc_insertion_point(class_scope:ExampleCompatDBModel)
  })
_sym_db.RegisterMessage(ExampleCompatDBModel)
_sym_db.RegisterMessage(ExampleCompatDBModel.MapStringStringEntry)
_sym_db.RegisterMessage(ExampleCompatDBModel.MapStringInt32Entry)


_EXAMPLECOMPATDBMODEL_MAPSTRINGSTRINGENTRY._options = None
_EXAMPLECOMPATDBMODEL_MAPSTRINGINT32ENTRY._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Copyright 2019 Tomaz Muraus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'GoogleDatastoreTranslatorIntegrationTestCase'
]

import sys

from google.cloud import datastore

from tests.generated import example_pb2
from tests.mocks import EXAMPLE_DICT_POPULATED
from tests.mocks import EXAMPLE_DICT_DEFAULT_VALUES
from tests.mocks import EXAMPLE_PB_POPULATED
from tests.mocks import EXAMPLE_PB_DEFAULT_VALUES
from tests.integration.base import BaseDatastoreIntegrationTestCase

from protobuf_cloud_datastore_translator import model_pb_to_entity_pb
from protobuf_cloud_datastore_translator import entity_pb_to_model_pb

__all__ = [
    'GoogleDatastoreTranslatorIntegrationTestCase'
]


class GoogleDatastoreTranslatorIntegrationTestCase(BaseDatastoreIntegrationTestCase):
    """
    NOTE: Those tests rely on datastore emulator running (gcloud beta emulator datastore start
    --no-store-on-disk).
    """

    def setUp(self):
        super(GoogleDatastoreTranslatorIntegrationTestCase, self).setUp()

        modules_to_remove = [
            'tests.generated.options_pb2',
            'tests.generated.models.options_pb2',
            'tests.generated.example_with_options_pb2',
            'tests.generated.models.example_with_options_pb2',
        ]

        for module_name in modules_to_remove:
            if module_name in sys.modules:
                del sys.modules[module_name]

    def test_store_and_retrieve_populated_translated_object_from_datastore(self):
        # type: () -> None
        """
        Test case which stores raw entity object in the datastore and verifies it matched the
        same object which is stored using translated Protobuf definition.
        """
        key_native = self.client.key('ExampleModel', 'native_entity_populated')

        entity_native = datastore.Entity(key=key_native)
        entity_native.update(EXAMPLE_DICT_POPULATED)
        self.client.put(entity_native)

        entity_native_retrieved = self.client.get(key_native)
        self.assertTrue(entity_native_retrieved)

        # Verify retrieved data matches the original input
        self.assertEqual(entity_native_retrieved, EXAMPLE_DICT_POPULATED)

        # Store custom Protobuf object in a datastore by translating it to Entity object
        key_translated = self.client.key('ExampleModel', 'translated_entity_populated')
        example_pb = EXAMPLE_PB_POPULATED
        entity_pb_translated = model_pb_to_entity_pb(model_pb=example_pb)

        # pylint: disable=no-member
        entity_pb_translated.key.CopyFrom(key_translated.to_protobuf())
        entity_translated = datastore.helpers.entity_from_protobuf(entity_pb_translated)
        self.client.put(entity_translated)

        # Verify that the translated entity results in the same end result as using native
        # entity object
        entity_translated_retrieved = self.client.get(key_translated)

        self.assertTrue(entity_translated_retrieved.key != entity_native_retrieved.key)

        # NOTE: key won't be the same so we clear it
        entity_translated_retrieved.key = None
        entity_native_retrieved.key = None

        self.assertEqual(entity_translated_retrieved, entity_native_retrieved)

        # If we translate retrieved entity back to the original Protobuf object definition, it
        # should be the same as the original model (minus the key since the original model doesn't
        # contain a key)
        entity_pb_retrieved = datastore.helpers.entity_to_protobuf(entity_translated_retrieved)
        entity_pb_translated.ClearField('key')
        self.assertEqual(entity_pb_translated, entity_pb_retrieved)

        example_pb_retrieved = entity_pb_to_model_pb(example_pb2.ExampleDBModel,
                                                     entity_pb_retrieved)
        self.assertEqual(example_pb_retrieved, example_pb)

    def test_store_and_retrieve_default_values_and_translated_object_from_datastore(self):
        # type: () -> None
        key_native = self.client.key('ExampleModel', 'native_entity_default_values')

        entity_native = datastore.Entity(key=key_native)
        entity_native.update(EXAMPLE_DICT_DEFAULT_VALUES)
        self.client.put(entity_native)

        entity_native_retrieved = self.client.get(key_native)
        self.assertTrue(entity_native_retrieved)

        # Verify retrieved data matches the original input
        self.assertEqual(entity_native_retrieved, EXAMPLE_DICT_DEFAULT_VALUES)

        # Store custom Protobuf object in a datastore by translating it to Entity object
        key_translated = self.client.key('ExampleModel', 'translated_entity_default_values')
        example_pb = EXAMPLE_PB_DEFAULT_VALUES
        entity_pb_translated = model_pb_to_entity_pb(model_pb=example_pb)
        # pylint: disable=no-member
        entity_pb_translated.key.CopyFrom(key_translated.to_protobuf())
        entity_translated = datastore.helpers.entity_from_protobuf(entity_pb_translated)
        self.client.put(entity_translated)

        # Verify that the translated entity results in the same end result as using native
        # entity object
        entity_translated_retrieved = self.client.get(key_translated)

        self.assertTrue(entity_translated_retrieved.key != entity_native_retrieved.key)

        # NOTE: key won't be the same so we clear it
        entity_translated_retrieved.key = None
        entity_native_retrieved.key = None

        self.assertEqual(entity_translated_retrieved, entity_native_retrieved)

        # If we translate retrieved entity back to the original Protobuf object definition, it
        # should be the same as the original model (minus the key since the original model doesn't
        # contain a key)
        entity_pb_retrieved = datastore.helpers.entity_to_protobuf(entity_translated_retrieved)
        entity_pb_translated.ClearField('key')
        self.assertEqual(entity_pb_translated, entity_pb_retrieved)

        example_pb_retrieved = entity_pb_to_model_pb(example_pb2.ExampleDBModel,
                                                     entity_pb_retrieved)
        self.assertEqual(example_pb_retrieved, example_pb)

        # Storing and retrieving empty object should have the same end result
        key_native_empty = self.client.key('ExampleModel', 'native_entity_empty')

        entity_native_empty = datastore.Entity(key=key_native_empty)
        entity_native_empty.update({})
        self.client.put(entity_native_empty)

        entity_native_empty_retrieved = self.client.get(key_native_empty)
        self.assertTrue(entity_native_empty_retrieved is not None)

        # Verify retrieved data matches the original input
        self.assertEqual(entity_native_empty_retrieved, {})

        # Store custom Protobuf object in a datastore by translating it to Entity object
        key_translated_empty = self.client.key('ExampleModel', 'translated_entity_empty')
        example_pb = example_pb2.ExampleDBModel()
        entity_pb_translated_empty = model_pb_to_entity_pb(model_pb=example_pb)
        # pylint: disable=no-member
        entity_pb_translated_empty.key.CopyFrom(key_translated_empty.to_protobuf())
        entity_translated_empty = datastore.helpers.entity_from_protobuf(entity_pb_translated_empty)
        self.client.put(entity_translated_empty)

        # Verify that the translated entity results in the same end result as using native
        # entity object
        entity_translated_empty_retrieved = self.client.get(key_translated_empty)

        self.assertTrue(entity_translated_empty_retrieved.key != entity_native_empty_retrieved.key)

        # NOTE: key won't be the same so we clear it
        entity_translated_empty_retrieved.key = None
        entity_native_empty_retrieved.key = None

        # self.assertEqual(entity_translated_empty_retrieved, entity_native_empty_retrieved)
        # return

        # If we translate retrieved entity back to the original Protobuf object definition, it
        # should be the same as the original model (minus the key since the original model doesn't
        # contain a key)
        entity_pb_empty_retrieved = \
            datastore.helpers.entity_to_protobuf(entity_translated_empty_retrieved)
        entity_pb_translated_empty.ClearField('key')
        entity_pb_empty_retrieved.ClearField('key')

        self.assertEqual(entity_pb_translated_empty, entity_pb_empty_retrieved)

        example_pb_empty_retrieved = entity_pb_to_model_pb(example_pb2.ExampleDBModel,
                                                           entity_pb_empty_retrieved)
        self.assertEqual(example_pb_empty_retrieved, example_pb)

    def test_model_pb_to_entity_pb_exclude_from_index_fields(self):
        # type: () -> None
        example_pb = example_pb2.ExampleDBModel()
        example_pb.int32_key = 100
        example_pb.string_key = 'string bar'
        example_pb.bytes_key = b'foobarbytes'
        example_pb.enum_key = 1  # type: ignore

        # No exclude from index provided
        entity_pb = model_pb_to_entity_pb(model_pb=example_pb)

        entity_translated = datastore.helpers.entity_from_protobuf(entity_pb)
        self.assertEqual(entity_translated.exclude_from_indexes, set([]))

        entity_translated.key = self.client.key('ExampleModel', 'exclude_from_indexes_1')
        self.client.put(entity_translated)

        entity_translated_retrieved = self.client.get(entity_translated.key)
        self.assertEqual(entity_translated, entity_translated_retrieved)

        # Exclude from index provided for some fields
        entity_pb = model_pb_to_entity_pb(model_pb=example_pb,
                                          exclude_from_index=['int32_key', 'bytes_key'])

        entity_translated = datastore.helpers.entity_from_protobuf(entity_pb)
        self.assertEqual(entity_translated.exclude_from_indexes, set(['int32_key', 'bytes_key']))

        entity_translated.key = self.client.key('ExampleModel', 'exclude_from_indexes_2')
        self.client.put(entity_translated)

        entity_translated_retrieved = self.client.get(entity_translated.key)
        self.assertEqual(entity_translated, entity_translated_retrieved)

    def test_model_pb_to_entity_pb_exclude_from_index_custom_extension_model_without_package(self):
        # type: () -> None
        from tests.generated import example_with_options_pb2

        model_pb = example_with_options_pb2.ExampleDBModelWithOptions1()
        model_pb.string_key_one = 'one'
        model_pb.string_key_two = 'two'
        model_pb.string_key_three = 'three'
        model_pb.string_key_four = 'four'
        model_pb.int32_field_one = 111
        model_pb.int32_field_two = 222

        entity_pb = model_pb_to_entity_pb(model_pb=model_pb)

        entity_translated = datastore.helpers.entity_from_protobuf(entity_pb)
        self.assertEqual(entity_translated.exclude_from_indexes,
                         set(['string_key_one', 'int32_field_two', 'string_key_three']))

        entity_translated.key = self.client.key('ExampleModelWithOptions', 'exclude_from_index_1')
        self.client.put(entity_translated)

        entity_translated_retrieved = self.client.get(entity_translated.key)
        self.assertEqual(entity_translated, entity_translated_retrieved)

    def test_model_pb_to_entity_pb_exclude_from_index_custom_extension_model_with_package(self):
        # type: () -> None
        from tests.generated.models import example_with_options_pb2

        model_pb = example_with_options_pb2.ExampleDBModelWithOptions1()
        model_pb.string_key_one = 'one'
        model_pb.string_key_two = 'two'
        model_pb.string_key_three = 'three'
        model_pb.string_key_four = 'four'
        model_pb.int32_field_one = 111
        model_pb.int32_field_two = 222

        entity_pb = model_pb_to_entity_pb(model_pb=model_pb)

        entity_translated = datastore.helpers.entity_from_protobuf(entity_pb)
        self.assertEqual(entity_translated.exclude_from_indexes,
                         set(['string_key_one', 'int32_field_two', 'string_key_three']))

        entity_translated.key = self.client.key('ExampleModelWithOptions', 'exclude_from_index_1')
        self.client.put(entity_translated)

        entity_translated_retrieved = self.client.get(entity_translated.key)
        self.assertEqual(entity_translated, entity_translated_retrieved)



# -*- coding: utf-8 -*-
# Copyright 2019 Tomaz Muraus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple HTTP server which allows users to retrieve and store arbitrary Protobuf objects inside Cloud
Datastore.

To run it:

$ PYTHONPATH=. gunicorn examples/http_api.py:app
"""

from typing import Tuple
from typing import Dict

from flask import Flask
from flask import request

from google.protobuf import json_format
from google.cloud import datastore

from protobuf_cloud_datastore_translator import model_pb_to_entity_pb
from protobuf_cloud_datastore_translator import entity_pb_to_model_pb
from protobuf_cloud_datastore_translator.utils import get_module_and_class_for_model_name

app = Flask(__name__)


@app.route('/datastore/put/<key>', methods=['POST'])
def put_db_object(key):
    # type: (str) -> Tuple[str, int, Dict[str, str]]
    """
    Store arbitrary Protobuf object in Google Datastore.

    NOTE: Request body needs to contain Protobuf model serialized as JSON.
    """
    body = request.get_json()

    # Fully qualified model name, e.g. "tests.generated.example_pb2.ExampleDBModel"
    # NOTE: This module needs to be available in PYTHONPATH
    model_name = body['model_name']
    model_data = body['json_string']

    _, model_class = get_module_and_class_for_model_name(model_name=model_name)

    model_pb = json_format.Parse(model_data, model_class())

    # 2. Convert it into entity object
    entity_pb = model_pb_to_entity_pb(model_pb)

    client = datastore.Client()

    # Set PK on the object
    key_pb = client.key(model_pb.DESCRIPTOR.name, key).to_protobuf()
    entity_pb.key.CopyFrom(key_pb)  # pylint: disable=no-member

    # 3. Store it inside datastore
    entity = datastore.helpers.entity_from_protobuf(entity_pb)
    client.put(entity)
    return '', 200, {}


@app.route('/datastore/get/<key>')
def get_db_object(key):
    # type: (str) -> Tuple[str, int, Dict[str, str]]
    """
    Retrieve object from Google Datastore, serialize it into native object type and serialize it
    as JSON.
    """
    model_name = request.args.get('model_name', '')
    raw = request.args.get('raw', 'false').lower() in ['1', 'true', 'yes']
    _, model_class = get_module_and_class_for_model_name(model_name=model_name)

    class_name = model_class.DESCRIPTOR.name

    # 1. Retrieve Entity from datastore
    client = datastore.Client()

    key = client.key(class_name, key)
    entity = client.get(key)

    # 2. Translate it to custom Protobuf object
    entity_pb = datastore.helpers.entity_to_protobuf(entity)

    if not raw:
        model_pb = entity_pb_to_model_pb(model_pb_class=model_class, entity_pb=entity_pb)
    else:
        model_pb = entity_pb

    # 3. Serialize it to JSON
    model_pb_json = json_format.MessageToJson(model_pb)

    return model_pb_json, 200, {'Content-Type': 'application/json'}



# -*- coding: utf-8 -*-
# Copyright 2019 Tomaz Muraus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import unittest

import requests
from google.cloud import datastore

from tests.mocks import EmulatorCreds

__all__ = [
    'BaseDatastoreIntegrationTestCase'
]

START_EMULATOR_STRING = """
gcloud beta emulators datastore start --host-port=127.0.0.1:8081 --no-store-on-disk --consistency=1
""".strip()


class BaseDatastoreIntegrationTestCase(unittest.TestCase):

    def setUp(self):
        # type: () -> None
        super(BaseDatastoreIntegrationTestCase, self).setUp()

        # Set environment variables which are needed for emulator to work
        os.environ['DATASTORE_DATASET'] = 'translator-tests'
        os.environ['DATASTORE_PROJECT_ID'] = 'translator-tests'
        os.environ['DATASTORE_EMULATOR_HOST'] = 'localhost:8081'
        os.environ['DATASTORE_EMULATOR_HOST_PATH'] = 'localhost:8081/datastore'
        os.environ['DATASTORE_HOST'] = 'http://localhost:8081'

        # 1. Verify datastore emulator is running
        try:
            requests.get(os.environ['DATASTORE_HOST'], timeout=1)
        except requests.exceptions.ConnectionError as e:
            raise ValueError('Can\'t reach "%s". Make sure Google Cloud Datastore emulator is '
                    'running and listening on "%s": %s.\n\nYou can start emulator using "%s" '
                    'command.' % (os.environ['DATASTORE_HOST'],
                                  os.environ['DATASTORE_EMULATOR_HOST'], str(e),
                                  START_EMULATOR_STRING))

        # Instantiate client with mock credentials object
        self.client = datastore.Client(credentials=EmulatorCreds(),
                _http=requests.Session())
        self._clear_datastore()

    def tearDown(self):
        # type: () -> None
        super(BaseDatastoreIntegrationTestCase, self).tearDown()

        self._clear_datastore()

    def _clear_datastore(self):
        # type: () -> None
        # Clear datastore, ensure it's empty
        query = self.client.query(kind='__kind__')
        query.keys_only()

        kinds = [entity.key.id_or_name for entity in query.fetch()]

        for kind in kinds:
            query = self.client.query(kind=kind)
            query.keys_only()

            # Work around for eventual consistency nature of the emulator
            for index in range(0, 3):
                entity_keys = [entity.key for entity in query.fetch()]
                self.client.delete_multi(entity_keys)
                time.sleep(0.1)

            query = self.client.query(kind=kind)
            query.keys_only()
            result = list(query.fetch())

            self.assertEqual(len(result), 0)



# -*- coding: utf-8 -*-
# Copyright 2019 Tomaz Muraus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib

from typing import Type
from typing import List
from typing import Tuple
from types import ModuleType

from google.protobuf.pyext.cpp_message import GeneratedProtocolMessageType

from protobuf_cloud_datastore_translator.translator import exclude_field_from_index


__all__ = [
    'get_module_and_class_for_model_name',
    'get_exclude_from_index_fields_for_model'
]


def get_module_and_class_for_model_name(model_name):
    # type: (str) -> Tuple[ModuleType, Type[GeneratedProtocolMessageType]]
    split = model_name.rsplit('.', 1)

    if len(split) != 2:
        raise ValueError('Invalid module name: %s' % (model_name))

    module_path, class_name = split

    try:
        module = importlib.import_module(module_path)
        model_class = getattr(module, class_name, None)
    except Exception as e:
        raise ValueError('Class "%s" not found: %s. Make sure "%s" is in PYTHONPATH' %
                         (model_name, module_path, str(e)))

    if not model_class:
        raise ValueError('Class "%s" not found in module "%s"' % (model_name, module_path))

    return module, model_class


def get_exclude_from_index_fields_for_model(model_class):
    # type: (Type[GeneratedProtocolMessageType]) -> List[str]
    """
    Return a list of fields which are marked as to be excluded for the provided model class.
    """

    fields = list(iter(model_class.DESCRIPTOR.fields))

    result = []
    for field_descriptor in fields:
        exclude_from_index = exclude_field_from_index(model=model_class,
                                                      field_descriptor=field_descriptor)

        if exclude_from_index:
            result.append(field_descriptor.name)

    return result



# -*- coding: utf-8 -*-
# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import re
import sys

from distutils.version import StrictVersion

# NOTE: This script can't rely on any 3rd party dependency so we need to use this code here
PY3 = sys.version_info[0] == 3

if PY3:
    text_type = str
else:
    text_type = unicode  # NOQA

GET_PIP = 'curl https://bootstrap.pypa.io/get-pip.py | python'

try:
    import pip
except ImportError as e:
    print('Failed to import pip: %s' % (text_type(e)))
    print('')
    print('Download pip:\n%s' % (GET_PIP))
    sys.exit(1)

__all__ = [
    'check_pip_version',
    'fetch_requirements',
    'apply_vagrant_workaround',
    'get_version_string',
    'parse_version_string'
]


def check_pip_version(min_version='6.0.0'):
    """
    Ensure that a minimum supported version of pip is installed.
    """
    if StrictVersion(pip.__version__) < StrictVersion(min_version):
        print("Upgrade pip, your version '{0}' "
              "is outdated. Minimum required version is '{1}':\n{2}".format(pip.__version__,
                                                                            min_version,
                                                                            GET_PIP))
        sys.exit(1)


def fetch_requirements(requirements_file_path):
    """
    Return a list of requirements and links by parsing the provided requirements file.
    """
    # TODO: Add support for links
    links = []
    reqs = []

    with open(requirements_file_path, 'r') as fp:
        for line in fp.readlines():
            line = line.strip()

            if line.startswith('#') or not line:
                continue

            reqs.append(line)

    return (reqs, links)


def apply_vagrant_workaround():
    """
    Function which detects if the script is being executed inside vagrant and if it is, it deletes
    "os.link" attribute.
    Note: Without this workaround, setup.py sdist will fail when running inside a shared directory
    (nfs / virtualbox shared folders).
    """
    if os.environ.get('USER', None) == 'vagrant':
        del os.link


def get_version_string(init_file):
    """
    Read __version__ string for an init file.
    """

    with open(init_file, 'r') as fp:
        content = fp.read()
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                  content, re.M)
        if version_match:
            return version_match.group(1)

        raise RuntimeError('Unable to find version string in %s.' % (init_file))


# alias for get_version_string
parse_version_string = get_version_string



# -*- coding: utf-8 -*-
# Copyright 2019 Tomaz Muraus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import os
import sys
import json
import uuid
import subprocess

from tests.integration.base import BaseDatastoreIntegrationTestCase

__all__ = [
    'CrossLangCompatibilityIntegrationTestCase'
]

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
FIXTURES_PATH = os.path.join(BASE_DIR, 'fixtures/')

PYTHON_INSERT_GET_SCRIPT_PATH = os.path.join(BASE_DIR, 'python/python-put-get-db-model.py')
GO_INSERT_GET_SCRIPT_PATH = os.path.join(BASE_DIR, 'go', 'go-put-get-db-model')

GO_BINARY_DOESNT_EXIST_ERROR = """
go-put-get-db-model binary doesn't exist. You can build it by running
./scripts/build-go-binary.sh script.
""".strip()


class CrossLangCompatibilityIntegrationTestCase(BaseDatastoreIntegrationTestCase):
    """
    Integration test which verifies that the output by Python and Go translator
    library is exactly the same.
    """

    maxDiff = None

    FIXTURES = [
        {
            'path': os.path.join(FIXTURES_PATH, 'example_db_model_1.json'),
            'key': str(uuid.uuid4()),
            'entity_kind': 'ExampleCompatDBModel'
        }
    ]

    def setUp(self):
        # type: () -> None
        super(CrossLangCompatibilityIntegrationTestCase, self).setUp()

        # Verify go binary exists
        if not os.path.isfile(GO_INSERT_GET_SCRIPT_PATH):
            raise ValueError(GO_BINARY_DOESNT_EXIST_ERROR)

        # Load fixture content into memory
        for fixture_obj in self.FIXTURES:
            with open(fixture_obj['path'], 'r') as fp:
                fixture_obj['content'] = json.loads(fp.read())

    def test_put_and_get(self):
        # type: () -> None
        for fixture_obj in self.FIXTURES:
            self._test_fixture_obj(fixture_obj=fixture_obj)

    def _test_fixture_obj(self, fixture_obj):
        # type: (Dict[str, str]) -> None

        # Insert Protobuf model in the datastore using Python translator library
        self._python_insert_fixture(fixture_obj=fixture_obj)

        # Insert Protobuf model in the datastore using Go translator library
        self._go_insert_fixture(fixture_obj=fixture_obj)

        # Verify entity has been inserted
        key_python = self.client.key(fixture_obj['entity_kind'], 'python_' + fixture_obj['key'])
        key_go = self.client.key(fixture_obj['entity_kind'], 'go_' + fixture_obj['key'])

        entity_python_pb = self.client.get(key_python)
        entity_go_pb = self.client.get(key_go)

        self.assertTrue(entity_python_pb, 'Entity with key "%s" not found' % (key_python))
        self.assertTrue(entity_go_pb, 'Entity with key "%s" not found' % (key_go))

        # Reset keys since they will always be different
        entity_python_pb.key = None
        entity_go_pb.key = None

        # Compare the raw entity result
        msg = 'Translated Entity PB objects for Python and Go don\'t match'
        self.assertEqual(entity_python_pb, entity_go_pb, msg)

        # Compare translated models

        # First perform a sanity test and make sure it matches the original fixture input
        model_pb_json_python = self._python_get_fixture(fixture_obj=fixture_obj)
        self.assertDictsEqualIgnoreMissingDefaultValues(model_pb_json_python,
                                                        fixture_obj['content'])

        model_pb_json_go = self._go_get_fixture(fixture_obj=fixture_obj)
        self.assertDictsEqualIgnoreMissingDefaultValues(model_pb_json_go, fixture_obj['content'])

        # Now compare Python and Go versions and make sure they match
        self.assertDictsEqualIgnoreMissingDefaultValues(model_pb_json_python, model_pb_json_go)

    def assertDictsEqualIgnoreMissingDefaultValues(self, model_pb_python, model_pb_go):
        """
        Custom assertion function which asserts that the provided Model PB returned by the Python
        translator library and the one returned by the Go one serialized as JSON are the same.

        NOTE: We need a custom assert functions because there are some differences between JSON
        <-> PB serialized in Python and Go and we use JSON as an intermediate format for our test
        fixtures.
        """
        field_names = set(model_pb_python.keys())
        field_names.update(model_pb_go.keys())

        for field_name in field_names:
            # NOTE: Due to the JSON serializer differences there can be some default values
            # missing
            value_python = model_pb_python.get(field_name, None)
            value_go = model_pb_go.get(field_name, None)

            if not value_python and not value_go:
                value_python = None
                value_go = None

            msg = 'Field "%s" on Python and Go Model PB object didn\'t match' % (field_name)
            self.assertEqual(value_python, value_go, msg)

    def _python_insert_fixture(self, fixture_obj):
        # type: (Dict[str, str]) -> None
        args = [
            sys.executable,
            PYTHON_INSERT_GET_SCRIPT_PATH,
            '--operation=put',
            '--fixture-path=%s' % (fixture_obj['path']),
            '--primary-key=python_%s' % (fixture_obj['key']),
        ]

        process = subprocess.Popen(args, shell=False, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        process.stdout, process.stderr = process.communicate()

        if process.returncode != 0:
            self.assertFalse('Failed to run command "%s": %s' % (args, process.stderr))

    def _go_insert_fixture(self, fixture_obj):
        # type: (Dict[str, str]) -> None
        args = [
            GO_INSERT_GET_SCRIPT_PATH,
            '-operation=put',
            '-fixture-path=%s' % (fixture_obj['path']),
            '-primary-key=go_%s' % (fixture_obj['key']),
        ]
        process = subprocess.Popen(args, shell=False, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        process.stdout, process.stderr = process.communicate()

        if process.returncode != 0:
            self.assertFalse('Failed to run command "%s": %s' % (args, process.stderr))

    def _python_get_fixture(self, fixture_obj):
        # type: (Dict[str, str]) -> Dict
        args = [
            sys.executable,
            PYTHON_INSERT_GET_SCRIPT_PATH,
            '--operation=get',
            '--primary-key=python_%s' % (fixture_obj['key']),
        ]
        process = subprocess.Popen(args, shell=False, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        process.stdout, process.stderr = process.communicate()

        if process.returncode != 0:
            self.assertFalse('Failed to run command "%s": %s' % (args, process.stderr))

        json_parsed = json.loads(process.stdout)  # type: ignore

        return json_parsed

    def _go_get_fixture(self, fixture_obj):
        # type: (Dict[str, str]) -> Dict
        args = [
            GO_INSERT_GET_SCRIPT_PATH,
            '-operation=get',
            '-primary-key=go_%s' % (fixture_obj['key']),
        ]
        process = subprocess.Popen(args, shell=False, stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        process.stdout, process.stderr = process.communicate()

        if process.returncode != 0:
            self.assertFalse('Failed to run command "%s": %s' % (args, process.stderr))

        json_parsed = json.loads(process.stdout)  # type: ignore

        return json_parsed



# -*- coding: utf-8 -*-
# Copyright 2019 Tomaz Muraus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=all

import sys
import unittest

import mock

from protobuf_cloud_datastore_translator.translator import get_python_module_for_field
from protobuf_cloud_datastore_translator.utils import get_module_and_class_for_model_name
from protobuf_cloud_datastore_translator.utils import get_exclude_from_index_fields_for_model
from tests.generated import example_pb2


class UtilsTestCase(unittest.TestCase):
    def test_get_module_and_class_for_model_name_success(self):
        model_name = 'tests.generated.example_pb2.ExampleDBModel'
        module, model_class = get_module_and_class_for_model_name(model_name)
        self.assertEqual(module, example_pb2)
        self.assertEqual(model_class, example_pb2.ExampleDBModel)

    def test_get_module_and_class_for_model_name_invalid_module_name(self):
        expected_msg = 'Class "some.not.found.Foo" not found'
        self.assertRaisesRegexp(ValueError, expected_msg, get_module_and_class_for_model_name,
                                'some.not.found.Foo')

        expected_msg = 'Invalid module name:'
        self.assertRaisesRegexp(ValueError, expected_msg, get_module_and_class_for_model_name,
                                'invalid')

    def test_get_module_and_class_for_model_name_invalid_class_name(self):
        model_name = 'tests.generated.example_pb2.Foo'
        expected_msg = 'Class "tests.generated.example_pb2.Foo" not found'
        self.assertRaisesRegexp(ValueError, expected_msg, get_module_and_class_for_model_name,
                                model_name)

    def test_get_python_module_for_field(self):
        module_name = 'example_pb2'
        full_module_name = 'tests.generated.example_pb2'

        # Module not in sys.module yet
        field = mock.Mock()
        field.message_type.file.name = module_name

        self._remove_module_from_sys_module(module_name)
        self._remove_module_from_sys_module(full_module_name)

        self.assertFalse(module_name in sys.modules)
        self.assertFalse(full_module_name in sys.modules)

        module = get_python_module_for_field(field=field)

        self.assertTrue(module)
        self.assertTrue(module_name in sys.modules)
        self.assertEqual(module.__name__, module_name)

        # Module already in sys.modules
        self.assertTrue(module_name in sys.modules)

        module = get_python_module_for_field(field=field)
        self.assertTrue(module)
        self.assertTrue(module_name in sys.modules)
        self.assertEqual(module.__name__, module_name)

        # Module already in sys.modules under an alias
        self._remove_module_from_sys_module(module_name)
        self._remove_module_from_sys_module(full_module_name)

        sys.modules[full_module_name] = module

        self.assertFalse(module_name in sys.modules)
        self.assertTrue(full_module_name in sys.modules)

        module = get_python_module_for_field(field=field)
        self.assertTrue(module)
        self.assertTrue(full_module_name in sys.modules)
        self.assertEqual(module.__name__, module_name)

    def test_get_python_module_for_field_invalid_module_name(self):
        field = mock.Mock()
        field.message_type.file.name = 'invalid.module'

        expected_msg = 'No module named'
        self.assertRaisesRegexp(ImportError, expected_msg,
                                get_python_module_for_field, field)

    def test_get_exclude_from_index_fields(self):
        from tests.generated import example_with_options_pb2

        model_class = example_with_options_pb2.ExampleDBModelWithOptions1
        exclude_fields = get_exclude_from_index_fields_for_model(model_class=model_class)

        self.assertEqual(exclude_fields, ['string_key_one', 'string_key_three', 'int32_field_two'])

        model_class = example_with_options_pb2.ExampleDBModelWithOptions2
        exclude_fields = get_exclude_from_index_fields_for_model(model_class=model_class)

        self.assertEqual(exclude_fields, ['int32_field_two'])

        model_class = example_with_options_pb2.ExampleDBModelWithOptions3
        exclude_fields = get_exclude_from_index_fields_for_model(model_class=model_class)

        self.assertEqual(exclude_fields, [])

    def _remove_module_from_sys_module(self, module_name):
        if module_name in sys.modules:
            del sys.modules[module_name]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: example.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.type import latlng_pb2 as google_dot_type_dot_latlng__pb2
import example2_pb2 as example2__pb2
from models import example3_pb2 as models_dot_example3__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='example.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\rexample.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x18google/type/latlng.proto\x1a\x0e\x65xample2.proto\x1a\x15models/example3.proto\"`\n\x12\x45xampleNestedModel\x12\x12\n\nstring_key\x18\x01 \x01(\t\x12\x11\n\tint32_key\x18\x02 \x01(\x05\x12#\n\x08\x65num_key\x18\x03 \x01(\x0e\x32\x11.ExampleEnumModel\"K\n\x15\x45xampleDBModelWithKey\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x12\n\nstring_key\x18\x02 \x01(\t\x12\x11\n\tint32_key\x18\x03 \x01(\x05\"\x93\x06\n\x0e\x45xampleDBModel\x12\x11\n\tint32_key\x18\x01 \x01(\x05\x12\x12\n\nstring_key\x18\x02 \x01(\t\x12\x10\n\x08\x62ool_key\x18\x03 \x01(\x08\x12\x11\n\tbytes_key\x18\x04 \x01(\x0c\x12\x12\n\ndouble_key\x18\x0e \x01(\x01\x12\x11\n\tfloat_key\x18\x0f \x01(\x02\x12\x11\n\tint64_key\x18\x10 \x01(\x03\x12?\n\x11map_string_string\x18\x05 \x03(\x0b\x32$.ExampleDBModel.MapStringStringEntry\x12=\n\x10map_string_int32\x18\x06 \x03(\x0b\x32#.ExampleDBModel.MapStringInt32Entry\x12\x18\n\x10string_array_key\x18\x07 \x03(\t\x12\x17\n\x0fint32_array_key\x18\x08 \x03(\x05\x12\x17\n\x0f\x62ytes_array_key\x18\x12 \x03(\x0c\x12.\n\x11\x63omplex_array_key\x18\t \x03(\x0b\x32\x13.ExampleNestedModel\x12\x31\n\x10struct_array_key\x18\x13 \x03(\x0b\x32\x17.google.protobuf.Struct\x12#\n\x08\x65num_key\x18\n \x01(\x0e\x32\x11.ExampleEnumModel\x12\x31\n\rtimestamp_key\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12+\n\nstruct_key\x18\x0c \x01(\x0b\x32\x17.google.protobuf.Struct\x12,\n\x08null_key\x18\r \x01(\x0e\x32\x1a.google.protobuf.NullValue\x12*\n\rgeo_point_key\x18\x11 \x01(\x0b\x32\x13.google.type.LatLng\x1a\x36\n\x14MapStringStringEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x35\n\x13MapStringInt32Entry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\"\x9f\x02\n ExampleWithReferencedTypeDBModel\x12\x12\n\nstring_key\x18\x01 \x01(\t\x12*\n\x0freferenced_enum\x18\x02 \x01(\x0e\x32\x11.ExampleEnumModel\x12\x33\n\x13referenced_type_key\x18\x03 \x01(\x0b\x32\x16.ExampleReferencedType\x12\x46\n\x1breferenced_package_type_key\x18\x04 \x01(\x0b\x32!.models.ExampleWithPackageDBModel\x12>\n\x15referenced_struct_key\x18\x05 \x01(\x0b\x32\x1f.ExampleWithNestedStructDBModel\"M\n\x1e\x45xampleWithNestedStructDBModel\x12+\n\nstruct_key\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct*3\n\x10\x45xampleEnumModel\x12\t\n\x05\x45NUM0\x10\x00\x12\t\n\x05\x45NUM1\x10\x01\x12\t\n\x05\x45NUM2\x10\x02\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_type_dot_latlng__pb2.DESCRIPTOR,example2__pb2.DESCRIPTOR,models_dot_example3__pb2.DESCRIPTOR,])

_EXAMPLEENUMMODEL = _descriptor.EnumDescriptor(
  name='ExampleEnumModel',
  full_name='ExampleEnumModel',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ENUM0', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENUM1', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ENUM2', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1479,
  serialized_end=1530,
)
_sym_db.RegisterEnumDescriptor(_EXAMPLEENUMMODEL)

ExampleEnumModel = enum_type_wrapper.EnumTypeWrapper(_EXAMPLEENUMMODEL)
ENUM0 = 0
ENUM1 = 1
ENUM2 = 2



_EXAMPLENESTEDMODEL = _descriptor.Descriptor(
  name='ExampleNestedModel',
  full_name='ExampleNestedModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='string_key', full_name='ExampleNestedModel.string_key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_key', full_name='ExampleNestedModel.int32_key', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='enum_key', full_name='ExampleNestedModel.enum_key', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=145,
  serialized_end=241,
)


_EXAMPLEDBMODELWITHKEY = _descriptor.Descriptor(
  name='ExampleDBModelWithKey',
  full_name='ExampleDBModelWithKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ExampleDBModelWithKey.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key', full_name='ExampleDBModelWithKey.string_key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_key', full_name='ExampleDBModelWithKey.int32_key', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=243,
  serialized_end=318,
)


_EXAMPLEDBMODEL_MAPSTRINGSTRINGENTRY = _descriptor.Descriptor(
  name='MapStringStringEntry',
  full_name='ExampleDBModel.MapStringStringEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ExampleDBModel.MapStringStringEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ExampleDBModel.MapStringStringEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=999,
  serialized_end=1053,
)

_EXAMPLEDBMODEL_MAPSTRINGINT32ENTRY = _descriptor.Descriptor(
  name='MapStringInt32Entry',
  full_name='ExampleDBModel.MapStringInt32Entry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ExampleDBModel.MapStringInt32Entry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ExampleDBModel.MapStringInt32Entry.value', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1055,
  serialized_end=1108,
)

_EXAMPLEDBMODEL = _descriptor.Descriptor(
  name='ExampleDBModel',
  full_name='ExampleDBModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='int32_key', full_name='ExampleDBModel.int32_key', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key', full_name='ExampleDBModel.string_key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bool_key', full_name='ExampleDBModel.bool_key', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bytes_key', full_name='ExampleDBModel.bytes_key', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='double_key', full_name='ExampleDBModel.double_key', index=4,
      number=14, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='float_key', full_name='ExampleDBModel.float_key', index=5,
      number=15, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int64_key', full_name='ExampleDBModel.int64_key', index=6,
      number=16, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='map_string_string', full_name='ExampleDBModel.map_string_string', index=7,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='map_string_int32', full_name='ExampleDBModel.map_string_int32', index=8,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_array_key', full_name='ExampleDBModel.string_array_key', index=9,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_array_key', full_name='ExampleDBModel.int32_array_key', index=10,
      number=8, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bytes_array_key', full_name='ExampleDBModel.bytes_array_key', index=11,
      number=18, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='complex_array_key', full_name='ExampleDBModel.complex_array_key', index=12,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='struct_array_key', full_name='ExampleDBModel.struct_array_key', index=13,
      number=19, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='enum_key', full_name='ExampleDBModel.enum_key', index=14,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp_key', full_name='ExampleDBModel.timestamp_key', index=15,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='struct_key', full_name='ExampleDBModel.struct_key', index=16,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='null_key', full_name='ExampleDBModel.null_key', index=17,
      number=13, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='geo_point_key', full_name='ExampleDBModel.geo_point_key', index=18,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_EXAMPLEDBMODEL_MAPSTRINGSTRINGENTRY, _EXAMPLEDBMODEL_MAPSTRINGINT32ENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=321,
  serialized_end=1108,
)


_EXAMPLEWITHREFERENCEDTYPEDBMODEL = _descriptor.Descriptor(
  name='ExampleWithReferencedTypeDBModel',
  full_name='ExampleWithReferencedTypeDBModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='string_key', full_name='ExampleWithReferencedTypeDBModel.string_key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='referenced_enum', full_name='ExampleWithReferencedTypeDBModel.referenced_enum', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='referenced_type_key', full_name='ExampleWithReferencedTypeDBModel.referenced_type_key', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='referenced_package_type_key', full_name='ExampleWithReferencedTypeDBModel.referenced_package_type_key', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='referenced_struct_key', full_name='ExampleWithReferencedTypeDBModel.referenced_struct_key', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1111,
  serialized_end=1398,
)


_EXAMPLEWITHNESTEDSTRUCTDBMODEL = _descriptor.Descriptor(
  name='ExampleWithNestedStructDBModel',
  full_name='ExampleWithNestedStructDBModel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='struct_key', full_name='ExampleWithNestedStructDBModel.struct_key', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1400,
  serialized_end=1477,
)

_EXAMPLENESTEDMODEL.fields_by_name['enum_key'].enum_type = _EXAMPLEENUMMODEL
_EXAMPLEDBMODEL_MAPSTRINGSTRINGENTRY.containing_type = _EXAMPLEDBMODEL
_EXAMPLEDBMODEL_MAPSTRINGINT32ENTRY.containing_type = _EXAMPLEDBMODEL
_EXAMPLEDBMODEL.fields_by_name['map_string_string'].message_type = _EXAMPLEDBMODEL_MAPSTRINGSTRINGENTRY
_EXAMPLEDBMODEL.fields_by_name['map_string_int32'].message_type = _EXAMPLEDBMODEL_MAPSTRINGINT32ENTRY
_EXAMPLEDBMODEL.fields_by_name['complex_array_key'].message_type = _EXAMPLENESTEDMODEL
_EXAMPLEDBMODEL.fields_by_name['struct_array_key'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_EXAMPLEDBMODEL.fields_by_name['enum_key'].enum_type = _EXAMPLEENUMMODEL
_EXAMPLEDBMODEL.fields_by_name['timestamp_key'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXAMPLEDBMODEL.fields_by_name['struct_key'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_EXAMPLEDBMODEL.fields_by_name['null_key'].enum_type = google_dot_protobuf_dot_struct__pb2._NULLVALUE
_EXAMPLEDBMODEL.fields_by_name['geo_point_key'].message_type = google_dot_type_dot_latlng__pb2._LATLNG
_EXAMPLEWITHREFERENCEDTYPEDBMODEL.fields_by_name['referenced_enum'].enum_type = _EXAMPLEENUMMODEL
_EXAMPLEWITHREFERENCEDTYPEDBMODEL.fields_by_name['referenced_type_key'].message_type = example2__pb2._EXAMPLEREFERENCEDTYPE
_EXAMPLEWITHREFERENCEDTYPEDBMODEL.fields_by_name['referenced_package_type_key'].message_type = models_dot_example3__pb2._EXAMPLEWITHPACKAGEDBMODEL
_EXAMPLEWITHREFERENCEDTYPEDBMODEL.fields_by_name['referenced_struct_key'].message_type = _EXAMPLEWITHNESTEDSTRUCTDBMODEL
_EXAMPLEWITHNESTEDSTRUCTDBMODEL.fields_by_name['struct_key'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
DESCRIPTOR.message_types_by_name['ExampleNestedModel'] = _EXAMPLENESTEDMODEL
DESCRIPTOR.message_types_by_name['ExampleDBModelWithKey'] = _EXAMPLEDBMODELWITHKEY
DESCRIPTOR.message_types_by_name['ExampleDBModel'] = _EXAMPLEDBMODEL
DESCRIPTOR.message_types_by_name['ExampleWithReferencedTypeDBModel'] = _EXAMPLEWITHREFERENCEDTYPEDBMODEL
DESCRIPTOR.message_types_by_name['ExampleWithNestedStructDBModel'] = _EXAMPLEWITHNESTEDSTRUCTDBMODEL
DESCRIPTOR.enum_types_by_name['ExampleEnumModel'] = _EXAMPLEENUMMODEL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ExampleNestedModel = _reflection.GeneratedProtocolMessageType('ExampleNestedModel', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLENESTEDMODEL,
  '__module__' : 'example_pb2'
  # @@protoc_insertion_point(class_scope:ExampleNestedModel)
  })
_sym_db.RegisterMessage(ExampleNestedModel)

ExampleDBModelWithKey = _reflection.GeneratedProtocolMessageType('ExampleDBModelWithKey', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLEDBMODELWITHKEY,
  '__module__' : 'example_pb2'
  # @@protoc_insertion_point(class_scope:ExampleDBModelWithKey)
  })
_sym_db.RegisterMessage(ExampleDBModelWithKey)

ExampleDBModel = _reflection.GeneratedProtocolMessageType('ExampleDBModel', (_message.Message,), {

  'MapStringStringEntry' : _reflection.GeneratedProtocolMessageType('MapStringStringEntry', (_message.Message,), {
    'DESCRIPTOR' : _EXAMPLEDBMODEL_MAPSTRINGSTRINGENTRY,
    '__module__' : 'example_pb2'
    # @@protoc_insertion_point(class_scope:ExampleDBModel.MapStringStringEntry)
    })
  ,

  'MapStringInt32Entry' : _reflection.GeneratedProtocolMessageType('MapStringInt32Entry', (_message.Message,), {
    'DESCRIPTOR' : _EXAMPLEDBMODEL_MAPSTRINGINT32ENTRY,
    '__module__' : 'example_pb2'
    # @@protoc_insertion_point(class_scope:ExampleDBModel.MapStringInt32Entry)
    })
  ,
  'DESCRIPTOR' : _EXAMPLEDBMODEL,
  '__module__' : 'example_pb2'
  # @@protoc_insertion_point(class_scope:ExampleDBModel)
  })
_sym_db.RegisterMessage(ExampleDBModel)
_sym_db.RegisterMessage(ExampleDBModel.MapStringStringEntry)
_sym_db.RegisterMessage(ExampleDBModel.MapStringInt32Entry)

ExampleWithReferencedTypeDBModel = _reflection.GeneratedProtocolMessageType('ExampleWithReferencedTypeDBModel', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLEWITHREFERENCEDTYPEDBMODEL,
  '__module__' : 'example_pb2'
  # @@protoc_insertion_point(class_scope:ExampleWithReferencedTypeDBModel)
  })
_sym_db.RegisterMessage(ExampleWithReferencedTypeDBModel)

ExampleWithNestedStructDBModel = _reflection.GeneratedProtocolMessageType('ExampleWithNestedStructDBModel', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLEWITHNESTEDSTRUCTDBMODEL,
  '__module__' : 'example_pb2'
  # @@protoc_insertion_point(class_scope:ExampleWithNestedStructDBModel)
  })
_sym_db.RegisterMessage(ExampleWithNestedStructDBModel)


_EXAMPLEDBMODEL_MAPSTRINGSTRINGENTRY._options = None
_EXAMPLEDBMODEL_MAPSTRINGINT32ENTRY._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Copyright 2019 Tomaz Muraus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import importlib

from typing import Any
from typing import Type
from typing import cast
from typing import Optional
from typing import Union
from typing import List
from typing import TypeVar
from types import ModuleType
from datetime import datetime

import six

from google.type import latlng_pb2
from google.cloud import datastore
from google.cloud.datastore_v1.proto import entity_pb2
from google.cloud.datastore.helpers import GeoPoint
from google.cloud.datastore_v1.types import Value
from google.protobuf import message
from google.protobuf import timestamp_pb2
from google.protobuf import struct_pb2
from google.protobuf import descriptor
from google.protobuf.internal.well_known_types import _GetStructValue  # type: ignore
from google.protobuf.pyext.cpp_message import GeneratedProtocolMessageType  # NOQA

from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.pyext._message import MessageMapContainer
from google.protobuf.pyext._message import ScalarMapContainer
from google.protobuf.pyext._message import RepeatedScalarContainer
from google.protobuf.pyext._message import RepeatedCompositeContainer

__all__ = [
    'model_pb_to_entity_pb',
    'model_pb_with_key_to_entity_pb',
    'entity_pb_to_model_pb'
]

# Type which represents an arbitrary ModelPB class which is a subclass of message.Message
T_model_pb = TypeVar('T_model_pb', bound=message.Message)

U_model_class_or_instance = Union[message.Message, Type[GeneratedProtocolMessageType]]

# String name for exclude from index extension which signals this library which model
# fields should be excluded from index
EXCLUDE_FROM_INDEX_EXT_NAME = 'exclude_from_index'


def model_pb_with_key_to_entity_pb(client, model_pb, exclude_falsy_values=False,
                                   exclude_from_index=None):
    # type: (datastore.Client, message.Message, bool, Optional[List[str]]) -> entity_pb2.Entity
    """
    Same as "model_pb_to_entity_pb", but it assumes model_pb which is passed to this function also
    contains "key" string field which is used to construct a primary key for the Entity PB object.

    NOTE: Datastore client instance needs to be passed to this method so
    namespace and project can be inferred from it (namespace_id and project_id are used as part of
    a composite primary key).
    """
    entity_pb = model_pb_to_entity_pb(model_pb=model_pb, exclude_falsy_values=exclude_falsy_values,
                                      exclude_from_index=exclude_from_index)

    if getattr(model_pb, 'key', None) is not None:
        # Special handling for top level key attribute which we assume will service as a primary
        # key (if provided)
        # NOTE: We use model name as the value for "kind" part of the key. Aka if Protobuf
        # message name is "MyClassDBModel", kind will be set to "MyClassDBModel"
        model_name = model_pb.DESCRIPTOR.name

        key_str = model_pb.key  # type: ignore
        key_pb = client.key(model_name, key_str).to_protobuf()
        entity_pb.key.CopyFrom(key_pb)  # pylint: disable=no-member

    return entity_pb


def model_pb_to_entity_pb(model_pb, exclude_falsy_values=False, exclude_from_index=None):
    # type: (message.Message, bool, Optional[List[str]]) -> entity_pb2.Entity
    """
    Translate Protobuf based database model object to Entity object which can be used with Google
    Datastore client library.

    :param model_pb: Instance of a custom Protobuf object to translate.

    :param exclude_falsy_values: True to exclude field values which are falsy (e.g. None, False,
                                 '', 0, etc.) and match the default values.

                                 NOTE: Due to the design of protobuf v3, there is no way to
                                 distinguish between a user explicitly providing a value which is
                                 the same as a default value (e.g. 0 for an integer field) and
                                 user not providing a value and default value being used instead.

    :param exclude_from_index: Optional list of field names which should not be indexed. By
                               default, all the simple fields are indexed.

                               NOTE: If provided, this value has high precedence over
                               "exclude_from_index" message option defined on the model.
    """
    exclude_from_index = exclude_from_index or []

    if not isinstance(model_pb, message.Message):
        raise ValueError('model_pb argument is not a valid Protobuf class instance')

    fields = list(iter(model_pb.DESCRIPTOR.fields))
    fields = [field for field in fields if field not in ['key']]

    entity_pb = entity_pb2.Entity()

    exclude_from_index = cast(list, exclude_from_index)

    for field_descriptor in fields:
        field_type = field_descriptor.type
        field_name = field_descriptor.name
        field_value = getattr(model_pb, field_name, None)

        if field_value is None:
            # Value not set or it uses a default value, skip it
            # NOTE: proto3 syntax doesn't support HasField() anymore so there is now way for us to
            # determine if a value is set / provided so we just use and return default values.
            continue

        if exclude_falsy_values and not field_value:
            continue

        attr_type = get_pb_attr_type(field_value)

        value_pb = None
        if attr_type == 'array_value':
            if len(field_value) == 0:
                value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
                array_value = entity_pb2.ArrayValue(values=[])
                value_pb.array_value.CopyFrom(array_value)
            else:
                value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)

                for value in field_value:
                    if field_type == descriptor.FieldDescriptor.TYPE_MESSAGE:
                        # Nested message type
                        entity_pb_item = model_pb_to_entity_pb(value)
                        value_pb_item = entity_pb2.Value()

                        # pylint: disable=no-member
                        value_pb_item.entity_value.CopyFrom(entity_pb_item)
                        # pylint: enable=no-member
                    else:
                        # Simple type
                        value_pb_item = entity_pb2.Value()
                        value_pb_item = set_value_pb_item_value(value_pb=value_pb_item, value=value)

                    value_pb.array_value.values.append(value_pb_item)
        elif field_type == descriptor.FieldDescriptor.TYPE_STRING:
            value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
            value_pb.string_value = field_value
        elif field_type in [descriptor.FieldDescriptor.TYPE_DOUBLE,
                descriptor.FieldDescriptor.TYPE_FLOAT]:
            # NOTE: Datastore only supports double type so we map float to double
            value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
            value_pb.double_value = field_value
        elif field_type in [descriptor.FieldDescriptor.TYPE_INT32,
                descriptor.FieldDescriptor.TYPE_INT64]:
            value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
            value_pb.integer_value = field_value
        elif field_type == descriptor.FieldDescriptor.TYPE_ENUM:
            value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)

            if field_descriptor.enum_type.name == 'NullValue':
                # NULL value
                value_pb.null_value = struct_pb2.NULL_VALUE
            else:
                # Regular ENUM
                value_pb.integer_value = field_value
        elif field_type == descriptor.FieldDescriptor.TYPE_BOOL:
            value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
            value_pb.boolean_value = field_value
        elif field_type == descriptor.FieldDescriptor.TYPE_BYTES:
            value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)

            if isinstance(field_value, six.string_types):
                field_value = field_value.encode('utf-8')

            value_pb.blob_value = field_value
        elif field_type == descriptor.FieldDescriptor.TYPE_MESSAGE:
            # Complex type, convert to entity
            field_type = model_pb.DESCRIPTOR.fields_by_name[field_name]

            if field_type.message_type.full_name == 'google.protobuf.Timestamp':
                if str(field_value) == '':
                    # Value not set
                    # TODO: Include default empty value?
                    # value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
                    # value_pb.timestamp_value.CopyFrom(field_value)
                    continue

                value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
                value_pb.timestamp_value.CopyFrom(field_value)
            elif field_type.message_type.full_name == 'google.type.LatLng':
                if str(field_value) == '':
                    # Value not set
                    continue
                value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
                value_pb.geo_point_value.CopyFrom(field_value)
            elif isinstance(field_value, MessageMapContainer):
                # Nested dictionary on a struct, set a value directory on a passed in pb object
                # which is a parent Struct entity
                entity_pb_item = get_entity_pb_for_value(value=field_value)
                entity_pb.CopyFrom(entity_pb_item)
            elif isinstance(field_value, ScalarMapContainer):
                # Custom user defined type, recurse into it
                value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
                entity_pb_item = get_entity_pb_for_value(value=field_value)
                value_pb.entity_value.CopyFrom(entity_pb_item)
            elif field_type.message_type.full_name == 'google.protobuf.Struct':
                if not dict(field_value):
                    # Value not set, skip it
                    continue

                value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
                entity_pb_item = get_entity_pb_for_value(value=field_value)
                value_pb.entity_value.CopyFrom(entity_pb_item)
            else:
                # Nested type, potentially referenced from another Protobuf definition file
                value_pb = datastore.helpers._new_value_pb(entity_pb, field_name)
                entity_pb_item = model_pb_to_entity_pb(field_value)
                value_pb.entity_value.CopyFrom(entity_pb_item)
        else:
            raise ValueError('Unsupported field type for field "%s"' % (field_name))

        if not value_pb:
            continue

        value_pb = cast(Value, value_pb)

        # Determine if field should be excluded from index
        exclude_field_from_indexes = exclude_field_from_index(model=model_pb,
                field_descriptor=field_descriptor,
                exclude_from_index=exclude_from_index)

        if exclude_field_from_indexes:
            # Field should be excluded from the index, mark that on the Entity Value
            value_pb.exclude_from_indexes = True

    return entity_pb


def entity_pb_to_model_pb(model_pb_class,   # type: Type[T_model_pb]
                          entity_pb,        # type: entity_pb2.Entity
                          strict=False      # type: bool
                          ):
    # type: (...) -> T_model_pb
    """
    Translate Google Datastore Entity Protobuf object to Protobuf based database model object.

    :param model_pb_class: Protobuf class to convert the Entity object to.
    :param entity_pb: Entity Protobuf instance to convert to database model instance.
    :param strict: True to run in a strict mode and throw an exception if we encounter a field on
                   the database object which is not defined on the model definition.
    """
    model_pb_field_names = list(iter(model_pb_class.DESCRIPTOR.fields))
    model_pb_field_names = [field.name for field in model_pb_field_names if field not in ['key']]

    model_pb = model_pb_class()

    for prop_name, value_pb in datastore.helpers._property_tuples(entity_pb):
        value = datastore.helpers._get_value_from_value_pb(value_pb)

        # Field not defined on the model class
        if prop_name not in model_pb_field_names:
            if strict:
                msg = ('Database object contains field "%s" which is not defined on the database '
                       'model class "%s"' % (prop_name, model_pb.DESCRIPTOR.name))
                raise ValueError(msg)
            else:
                continue

        def set_model_pb_value(model_pb, prop_name, value, is_nested=False):
            model_pb_class = model_pb.__class__

            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        # Handle nested models
                        if model_pb_class.DESCRIPTOR.fields_by_name[prop_name].message_type:
                            field = model_pb_class.DESCRIPTOR.fields_by_name[prop_name]

                            # Dynamically import nested model from a corresponding file
                            nested_model_name = field.message_type.name
                            nested_model_module = get_python_module_for_field(field=field)
                            nested_model_class = getattr(nested_model_module, nested_model_name)

                            # Instantiate an instance of nested field Protobuf class
                            item_pb = nested_model_class()
                            set_model_pb_value(item_pb, prop_name, item, is_nested=True)

                            getattr(model_pb, prop_name).append(item_pb)
                    elif isinstance(model_pb, struct_pb2.Struct):
                        try:
                            model_pb[prop_name]
                        except ValueError:
                            model_pb.update({prop_name: []})

                        model_pb[prop_name].append(item)
                    else:
                        getattr(model_pb, prop_name).append(item)
            elif isinstance(value, dict):
                # We assume it's a referenced protobuf type if it doesn't contain "update()" method
                # google.protobuf.Struct and Map types contain "update()" methods so we can treat
                # them as simple dictionaries
                if is_nested:
                    for key, value in six.iteritems(value):
                        set_model_pb_value(model_pb, key, value)
                elif isinstance(model_pb, struct_pb2.Struct):
                    model_pb.update({prop_name: value})
                else:
                    field = model_pb_class.DESCRIPTOR.fields_by_name[prop_name]
                    is_nested_model_type = (bool(field.message_type) and
                                            not hasattr(getattr(model_pb, prop_name, {}), 'update'))

                    if is_nested_model_type:
                        # Custom type definition potentially defined in different file
                        field = model_pb_class.DESCRIPTOR.fields_by_name[prop_name]

                        # Dynamically import nested model from a corresponding file
                        nested_model_name = field.message_type.name
                        nested_model_module = get_python_module_for_field(field=field)
                        nested_model_class = getattr(nested_model_module, nested_model_name)

                        item_pb = nested_model_class()
                        set_model_pb_value(item_pb, prop_name, value, is_nested=True)

                        getattr(model_pb, prop_name).CopyFrom(item_pb)
                    else:
                        getattr(model_pb, prop_name).update(dict(value))
            elif isinstance(value, datetime):
                getattr(model_pb, prop_name).FromDatetime(value)
            elif value is None:
                # NULL type
                setattr(model_pb, prop_name, 0)
            elif isinstance(value, GeoPoint):
                item_pb = latlng_pb2.LatLng(latitude=value.latitude, longitude=value.longitude)
                getattr(model_pb, prop_name).CopyFrom(item_pb)
            elif isinstance(model_pb, struct_pb2.Struct):
                model_pb.update({prop_name: value})
            else:
                setattr(model_pb, prop_name, value)

        set_model_pb_value(model_pb, prop_name, value)

    return model_pb


def get_pb_attr_type(value):
    # type: (Any) -> str
    """
    Return protobuf attribute type for the provided Python or protobuf value.
    """
    if isinstance(value, timestamp_pb2.Timestamp):
        name = 'timestamp'
    elif isinstance(value, bool):
        name = 'boolean'
    elif isinstance(value, float):
        name = 'double'
    elif isinstance(value, six.integer_types):
        name = 'integer'
    elif isinstance(value, six.text_type):
        name = 'string'
    elif isinstance(value, six.binary_type):
        name = 'blob'
    elif isinstance(value, (dict, ScalarMapContainer, MessageMapContainer, struct_pb2.Struct,
                            message.Message)):
        name = 'dict'
    elif isinstance(value, (list, RepeatedScalarContainer, RepeatedCompositeContainer)):
        name = 'array'
    elif value is None:
        name = 'null'
    else:
        raise ValueError('Unknown protobuf attr type', type(value))

    return name + '_value'


def get_entity_pb_for_value(value):
    # type: (Any) -> entity_pb2.Entity
    """
    Return Entity protobuf object for the provided Python value.
    """
    entity_pb = entity_pb2.Entity()

    attr_type = get_pb_attr_type(value)

    if attr_type == 'dict_value':
        if six.PY2:
            value = dict(value)

        for key, value in six.iteritems(value):
            value_pb = datastore.helpers._new_value_pb(entity_pb, key)
            value_pb = set_value_pb_item_value(value_pb=value_pb, value=value)
    else:
        raise ValueError('Unsupported attribute type: %s' % (attr_type))

    return entity_pb


def set_value_pb_item_value(value_pb, value):
    # type: (entity_pb2.Value, Any) -> entity_pb2.Value
    """
    Set a value attribute on the Value object based on the type of the provided value.

    NOTE: For complex nested types (e.g. dicts and structs this function uses recursion).
    """
    if isinstance(value, struct_pb2.ListValue):
        # Cast special ListValue type to a list
        value = cast(Any, value)
        value = list(value)

    if isinstance(value, float) and value.is_integer():
        # Special case because of how Protobuf handles ints in some scenarios (e.g. Struct)
        value = cast(Any, value)
        value = int(value)

    if isinstance(value, six.text_type):
        value_pb.string_value = value
    elif isinstance(value, int):
        value_pb.integer_value = value
    elif isinstance(value, float):
        value_pb.double_value = value
    elif isinstance(value, six.binary_type):
        value_pb.blob_value = value
    elif isinstance(value, list):
        if len(value) == 0:
            array_value = entity_pb2.ArrayValue(values=[])
            value_pb.array_value.CopyFrom(array_value)
        else:
            for value in value:
                value_pb_item = entity_pb2.Value()
                value_pb_item = set_value_pb_item_value(value_pb=value_pb_item, value=value)

                value_pb.array_value.values.append(value_pb_item)
    elif isinstance(value, struct_pb2.Value):
        item_value = _GetStructValue(value)
        set_value_pb_item_value(value_pb, item_value)
    elif hasattr(value, 'DESCRIPTOR'):
        # Custom user-defined type
        entity_pb_item = model_pb_to_entity_pb(value, exclude_falsy_values=True)
        value_pb.entity_value.CopyFrom(entity_pb_item)
    elif value is None:
        value_pb.null_value = struct_pb2.NULL_VALUE
    else:
        raise ValueError('Unsupported type for value: %s' % (value))

    return value_pb


def exclude_field_from_index(model, field_descriptor, exclude_from_index=None):
    # type: (U_model_class_or_instance, FieldDescriptor, Optional[List[str]]) -> bool
    """
    Return True if a particular field should be excluded from index, False otherwise.

    :param model: Either a Protobuf model class type or actual Protobuf model class instance.
    """
    # Determine if field should be excluded from index based on the "exclude_from_index"
    # function argument value
    # NOTE: This value has precedence over field level option
    if exclude_from_index:
        if field_descriptor.name in exclude_from_index:
            return True
        else:
            return False

    # Determine if field should be excluded from index based on the custom
    # "exclude_from_index" field level option
    field_exts = field_descriptor.GetOptions().Extensions

    # Bail early to avoid unncessary extension processing if there are no extensions defined
    if len(field_exts) == 0:  # type: ignore
        return False

    exclude_from_index_ext = None

    # If model file is part of a package, try searching for extension inside the package first
    if model.DESCRIPTOR.file.package:
        ext_name = '%s.%s' % (model.DESCRIPTOR.file.package, EXCLUDE_FROM_INDEX_EXT_NAME)
        exclude_from_index_ext = field_exts._FindExtensionByName(ext_name)  # type: ignore

    # And if it's not found inside the package or the model is not part of a package, try to
    # search for it in a top level namespace
    if not exclude_from_index_ext:
        ext_name = EXCLUDE_FROM_INDEX_EXT_NAME
        exclude_from_index_ext = field_exts._FindExtensionByName(ext_name)  # type: ignore

    if not exclude_from_index_ext:
        # Exclude from index extension not found
        return False

    try:
        exclude_from_index_ext_value = field_exts[exclude_from_index_ext]
    except KeyError:
        return False

    if exclude_from_index_ext_value is True:
        return True

    return False


def get_python_module_for_field(field):
    # type: (FieldDescriptor) -> ModuleType
    """
    Return Python module for the provided Protobuf field.

    NOTE: This function will also import the module if it's not already available in sys.path.
    """
    model_file = field.message_type.file.name
    module_name = model_file.replace('.proto', '_pb2').replace('/', '.')

    module = None

    # Check if module is already loaded
    if module_name in sys.modules:
        # Module already in sys.modules under the same import name
        module = sys.modules[module_name]
    else:
        # Check if module is in sys.modules under a different import name aka alias
        for name in sys.modules:
            if name.endswith(module_name):
                module = sys.modules[name]
                break

    if not module:
        # Module not in sys.modules, import it
        module = importlib.import_module(module_name)

    return module



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: example_with_options.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
from . import options_pb2 as options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='example_with_options.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1a\x65xample_with_options.proto\x1a google/protobuf/descriptor.proto\x1a\roptions.proto\"\xc3\x01\n\x1a\x45xampleDBModelWithOptions1\x12\x1c\n\x0estring_key_one\x18\x01 \x01(\tB\x04\x80\xb5\x18\x01\x12\x16\n\x0estring_key_two\x18\x02 \x01(\t\x12\x1e\n\x10string_key_three\x18\x03 \x01(\tB\x04\x80\xb5\x18\x01\x12\x17\n\x0fstring_key_four\x18\x04 \x01(\t\x12\x17\n\x0fint32_field_one\x18\x05 \x01(\x05\x12\x1d\n\x0fint32_field_two\x18\x06 \x01(\x05\x42\x04\x80\xb5\x18\x01\"\xb7\x01\n\x1a\x45xampleDBModelWithOptions2\x12\x16\n\x0estring_key_one\x18\x01 \x01(\t\x12\x16\n\x0estring_key_two\x18\x02 \x01(\t\x12\x18\n\x10string_key_three\x18\x03 \x01(\t\x12\x17\n\x0fstring_key_four\x18\x04 \x01(\t\x12\x17\n\x0fint32_field_one\x18\x05 \x01(\x05\x12\x1d\n\x0fint32_field_two\x18\x06 \x01(\x05\x42\x04\x80\xb5\x18\x01\"\xb1\x01\n\x1a\x45xampleDBModelWithOptions3\x12\x16\n\x0estring_key_one\x18\x01 \x01(\t\x12\x16\n\x0estring_key_two\x18\x02 \x01(\t\x12\x18\n\x10string_key_three\x18\x03 \x01(\t\x12\x17\n\x0fstring_key_four\x18\x04 \x01(\t\x12\x17\n\x0fint32_field_one\x18\x05 \x01(\x05\x12\x17\n\x0fint32_field_two\x18\x06 \x01(\x05\"\xc8\x01\n!ExampleDBModelWithMultipleOptions\x12 \n\x0estring_key_one\x18\x01 \x01(\tB\x08\x80\xb5\x18\x01\x80\xa6\x1d\x01\x12\x16\n\x0estring_key_two\x18\x02 \x01(\t\x12\x18\n\x10string_key_three\x18\x03 \x01(\t\x12\x17\n\x0fstring_key_four\x18\x04 \x01(\t\x12\x1d\n\x0fint32_field_one\x18\x05 \x01(\x05\x42\x04\x80\xa6\x1d\x01\x12\x17\n\x0fint32_field_two\x18\x06 \x01(\x05:=\n\x14my_other_bool_option\x12\x1d.google.protobuf.FieldOptions\x18\xe0\xd4\x03 \x01(\x08\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,options__pb2.DESCRIPTOR,])


MY_OTHER_BOOL_OPTION_FIELD_NUMBER = 60000
my_other_bool_option = _descriptor.FieldDescriptor(
  name='my_other_bool_option', full_name='my_other_bool_option', index=0,
  number=60000, type=8, cpp_type=7, label=1,
  has_default_value=False, default_value=False,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR)


_EXAMPLEDBMODELWITHOPTIONS1 = _descriptor.Descriptor(
  name='ExampleDBModelWithOptions1',
  full_name='ExampleDBModelWithOptions1',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='string_key_one', full_name='ExampleDBModelWithOptions1.string_key_one', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\200\265\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_two', full_name='ExampleDBModelWithOptions1.string_key_two', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_three', full_name='ExampleDBModelWithOptions1.string_key_three', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\200\265\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_four', full_name='ExampleDBModelWithOptions1.string_key_four', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_field_one', full_name='ExampleDBModelWithOptions1.int32_field_one', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_field_two', full_name='ExampleDBModelWithOptions1.int32_field_two', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\200\265\030\001'), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=80,
  serialized_end=275,
)


_EXAMPLEDBMODELWITHOPTIONS2 = _descriptor.Descriptor(
  name='ExampleDBModelWithOptions2',
  full_name='ExampleDBModelWithOptions2',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='string_key_one', full_name='ExampleDBModelWithOptions2.string_key_one', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_two', full_name='ExampleDBModelWithOptions2.string_key_two', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_three', full_name='ExampleDBModelWithOptions2.string_key_three', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_four', full_name='ExampleDBModelWithOptions2.string_key_four', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_field_one', full_name='ExampleDBModelWithOptions2.int32_field_one', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_field_two', full_name='ExampleDBModelWithOptions2.int32_field_two', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\200\265\030\001'), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=278,
  serialized_end=461,
)


_EXAMPLEDBMODELWITHOPTIONS3 = _descriptor.Descriptor(
  name='ExampleDBModelWithOptions3',
  full_name='ExampleDBModelWithOptions3',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='string_key_one', full_name='ExampleDBModelWithOptions3.string_key_one', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_two', full_name='ExampleDBModelWithOptions3.string_key_two', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_three', full_name='ExampleDBModelWithOptions3.string_key_three', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_four', full_name='ExampleDBModelWithOptions3.string_key_four', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_field_one', full_name='ExampleDBModelWithOptions3.int32_field_one', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_field_two', full_name='ExampleDBModelWithOptions3.int32_field_two', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=464,
  serialized_end=641,
)


_EXAMPLEDBMODELWITHMULTIPLEOPTIONS = _descriptor.Descriptor(
  name='ExampleDBModelWithMultipleOptions',
  full_name='ExampleDBModelWithMultipleOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='string_key_one', full_name='ExampleDBModelWithMultipleOptions.string_key_one', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\200\265\030\001\200\246\035\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_two', full_name='ExampleDBModelWithMultipleOptions.string_key_two', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_three', full_name='ExampleDBModelWithMultipleOptions.string_key_three', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_key_four', full_name='ExampleDBModelWithMultipleOptions.string_key_four', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_field_one', full_name='ExampleDBModelWithMultipleOptions.int32_field_one', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\200\246\035\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int32_field_two', full_name='ExampleDBModelWithMultipleOptions.int32_field_two', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=644,
  serialized_end=844,
)

DESCRIPTOR.message_types_by_name['ExampleDBModelWithOptions1'] = _EXAMPLEDBMODELWITHOPTIONS1
DESCRIPTOR.message_types_by_name['ExampleDBModelWithOptions2'] = _EXAMPLEDBMODELWITHOPTIONS2
DESCRIPTOR.message_types_by_name['ExampleDBModelWithOptions3'] = _EXAMPLEDBMODELWITHOPTIONS3
DESCRIPTOR.message_types_by_name['ExampleDBModelWithMultipleOptions'] = _EXAMPLEDBMODELWITHMULTIPLEOPTIONS
DESCRIPTOR.extensions_by_name['my_other_bool_option'] = my_other_bool_option
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ExampleDBModelWithOptions1 = _reflection.GeneratedProtocolMessageType('ExampleDBModelWithOptions1', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLEDBMODELWITHOPTIONS1,
  '__module__' : 'example_with_options_pb2'
  # @@protoc_insertion_point(class_scope:ExampleDBModelWithOptions1)
  })
_sym_db.RegisterMessage(ExampleDBModelWithOptions1)

ExampleDBModelWithOptions2 = _reflection.GeneratedProtocolMessageType('ExampleDBModelWithOptions2', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLEDBMODELWITHOPTIONS2,
  '__module__' : 'example_with_options_pb2'
  # @@protoc_insertion_point(class_scope:ExampleDBModelWithOptions2)
  })
_sym_db.RegisterMessage(ExampleDBModelWithOptions2)

ExampleDBModelWithOptions3 = _reflection.GeneratedProtocolMessageType('ExampleDBModelWithOptions3', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLEDBMODELWITHOPTIONS3,
  '__module__' : 'example_with_options_pb2'
  # @@protoc_insertion_point(class_scope:ExampleDBModelWithOptions3)
  })
_sym_db.RegisterMessage(ExampleDBModelWithOptions3)

ExampleDBModelWithMultipleOptions = _reflection.GeneratedProtocolMessageType('ExampleDBModelWithMultipleOptions', (_message.Message,), {
  'DESCRIPTOR' : _EXAMPLEDBMODELWITHMULTIPLEOPTIONS,
  '__module__' : 'example_with_options_pb2'
  # @@protoc_insertion_point(class_scope:ExampleDBModelWithMultipleOptions)
  })
_sym_db.RegisterMessage(ExampleDBModelWithMultipleOptions)

google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(my_other_bool_option)

_EXAMPLEDBMODELWITHOPTIONS1.fields_by_name['string_key_one']._options = None
_EXAMPLEDBMODELWITHOPTIONS1.fields_by_name['string_key_three']._options = None
_EXAMPLEDBMODELWITHOPTIONS1.fields_by_name['int32_field_two']._options = None
_EXAMPLEDBMODELWITHOPTIONS2.fields_by_name['int32_field_two']._options = None
_EXAMPLEDBMODELWITHMULTIPLEOPTIONS.fields_by_name['string_key_one']._options = None
_EXAMPLEDBMODELWITHMULTIPLEOPTIONS.fields_by_name['int32_field_one']._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Copyright 2019 Tomaz Muraus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import sys
import os.path

from setuptools import setup
from setuptools import find_packages

from dist_utils import fetch_requirements
from dist_utils import parse_version_string

PY2 = sys.version_info[0] == 2
PY2_or_3_pre_34 = sys.version_info < (3, 4, 0)

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
REQUIREMENTS_FILE = os.path.join(BASE_DIR, 'requirements.txt')
INIT_FILE = os.path.join(BASE_DIR, 'protobuf_cloud_datastore_translator', '__init__.py')

version = parse_version_string(INIT_FILE)
install_reqs, dep_links = fetch_requirements(REQUIREMENTS_FILE)

if PY2_or_3_pre_34:
    install_reqs.append('typing')

with codecs.open(os.path.join(BASE_DIR, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='protobuf-cloud-datastore-translator',
    version=version,
    description=('Library which converts arbitrary Protobuf message objects into '
                 'Entity Protobuf objects which can be used with Google Datastore.'),
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Tomaz Muraus',
    author_email='tomaz@tomaz.me',
    license='Apache License (2.0)',
    classifiers=[
        'Intended Audience :: Information Technology',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=install_reqs,
    dependency_links=dep_links,
    zip_safe=False,
    include_package_data=True,
    packages=find_packages(exclude=['setuptools', 'tests']),
    package_data={
        'protobuf_cloud_datastore_translator': ['py.typed']
    }
)
